package com.framsticks.core;

import org.apache.log4j.Logger;

import com.framsticks.params.AccessInterface;
import com.framsticks.params.CompositeParam;
import com.framsticks.params.FramsClass;
import com.framsticks.params.Param;
import com.framsticks.params.PrimitiveParam;
import com.framsticks.params.ValueParam;
import com.framsticks.params.annotations.AutoAppendAnnotation;
import com.framsticks.params.annotations.FramsClassAnnotation;
import com.framsticks.params.types.ProcedureParam;
import com.framsticks.util.FramsticksException;
import com.framsticks.util.dispatching.Future;

import static com.framsticks.core.TreeOperations.*;

@FramsClassAnnotation
public class ObjectTree extends AbstractTree {
	private static final Logger log = Logger.getLogger(ObjectTree.class);

	@AutoAppendAnnotation
	public void setRootObject(Object object) {
		registry.registerAndBuild(object.getClass());
		AccessInterface access = registry.createAccess(object.getClass());
		assignRootParam(Param.build().forAccess(access).id(getName()).finish(CompositeParam.class));
		assignRootObject(object);
	}

	public Object getRootObject() {
		return getAssignedRoot().getObject();
	}

	public <T> T getRootObject(Class<T> type) {
		Object result = getRootObject();
		if (result == null) {
			throw new FramsticksException().msg("object tree is empty").arg("tree", this);
		}
		if (!type.isInstance(result)) {
			throw new FramsticksException().msg("object tree holds object of different kind").arg("object", result).arg("requested", type).arg("tree", this);
		}
		return type.cast(result);
	}

	@Override
	public void get(Path path, Mode mode, Future<Path> future) {
		assert isActive();
		log.debug("requesting: " + path);
		path = resolveTopSync(path);
		future.pass(path);
	}

	@Override
	public void get(Path path, ValueParam param, Mode mode, Future<Object> future) {
		assert isActive();
		path = resolveTopSync(path);
		future.pass(bindAccess(path).get(param, Object.class));
	}

	@Override
	public void call(Path path, ProcedureParam param, Object[] arguments, Future<Object> future) {
		assert isActive();
		try {
			future.pass(bindAccess(path).call(param, arguments));
		} catch (FramsticksException e) {
			future.handle(e);
		}
	}

	@Override
	public void info(Path path, Future<FramsClass> future) {
		assert isActive();
		Path p = path.tryResolveIfNeeded();
		Class<?> javaClass = p.getTopObject().getClass();
		FramsClass framsClass = registry.registerReflectedIfNeeded(javaClass);
		if (framsClass != null) {
			future.pass(framsClass);
		} else {
			future.handle(new FramsticksException().msg("failed to find info for class").arg("java class", javaClass));
		}
	}

	protected Path resolveTopSync(Path path) {
		assert isActive();
		assert path.isOwner(this);
		if (path.getTop().getObject() != null) {
			return path;
		}
		AccessInterface access = bindAccess(path.getUnder());
		Object object = access.get(path.getTop().getParam(), Object.class);
		if (object == null) {
			return path;
		}
		return path.appendResolution(object);
	}


	@Override
	public void set(Path path, PrimitiveParam<?> param, Object value, final Future<Integer> future) {
		assert isActive();
		future.pass(bindAccess(path).set(param, value));
	}

}
