package com.framsticks.gui;

import java.lang.ref.WeakReference;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

import javax.swing.tree.TreePath;

import org.apache.log4j.Logger;

import com.framsticks.core.Node;
import com.framsticks.core.Tree;
import com.framsticks.gui.view.TreeCellRenderer;
import com.framsticks.params.AccessInterface;
import com.framsticks.params.CompositeParam;
import com.framsticks.params.types.StringParam;
import com.framsticks.util.FramsticksException;
import com.framsticks.util.lang.Casting;
import com.framsticks.util.lang.Pair;
import com.framsticks.util.swing.TooltipConstructor;

public class TreeNode extends AbstractNode {
	private static final Logger log = Logger.getLogger(TreeNode.class);


	protected static final AtomicInteger counter = new AtomicInteger();

	protected final WeakReference<Object> reference;
	protected final CompositeParam param;
	protected final Tree tree;
	protected final List<Pair<WeakReference<Object>, WeakReference<TreeNode>>> children = new LinkedList<>();
	protected final int number;

	public TreeNode(Tree tree, CompositeParam param, Object object) {
		this.tree = tree;
		this.param = param;
		reference = new WeakReference<Object>(object);
		number = counter.getAndIncrement();
	}

	public TreeNode(Node node) {
		this(node.getTree(), node.getParam(), node.assureResolved().getObject());
	}

	public Node tryCreateNode() {
		Object child = lock();
		if (child == null) {
			return null;
		}
		return new Node(tree, param, child);
	}

	@Override
	public int getChildCount() {
		Object referent = lock();
		if (referent == null) {
			return 0;
		}
		AccessInterface access = tree.prepareAccess(param).select(referent);
		final int count = access.getCompositeParamCount();
		return count;
	}

	public TreeNode getTreeNodeForChild(Object child) {
		Iterator<Pair<WeakReference<Object>, WeakReference<TreeNode>>> i = children.iterator();
		while (i.hasNext()) {
			Pair<WeakReference<Object>, WeakReference<TreeNode>> p = i.next();
			Object object = p.first.get();
			if (object == null) {
				i.remove();
				continue;
			}
			TreeNode treeNode = p.second.get();
			if (treeNode == null) {
				i.remove();
				continue;
			}
			if (object == child) {
				return treeNode;
			}
		}
		return null;

		// WeakReference<GuiTreeNode> resultReference = children.get(child);
		// if (resultReference == null) {
		//	return null;
		// }
		// return resultReference.get();
	}

	@Override
	public AbstractNode getChild(int number) {
		Object referent = lock();
		if (referent == null) {
			throw new FramsticksException().msg("invalid state - missing referent");
		}
		AccessInterface access = tree.prepareAccess(param).select(referent);

		final int count = access.getCompositeParamCount();
		if (number >= count) {
			throw new FramsticksException().msg("invalid state - no child");
		}

		CompositeParam childParam = access.getCompositeParam(number);
		Object child = access.get(childParam, Object.class);
		if (child == null) {
			log.debug("returing dummy node for " + childParam + " in " + referent);
			return new EmptyNode(childParam);
		}

		TreeNode result = getTreeNodeForChild(child);
		if (result != null) {
			return result;
		}
		result = new TreeNode(tree, childParam, child);

		children.add(Pair.make(new WeakReference<Object>(child), new WeakReference<TreeNode>(result)));

		return result;

	}

	public Object lock() {
		return reference.get();
	}

	@Override
	public int getIndexOfChild(AbstractNode child) {
		final TreeNode treeChild = Casting.tryCast(TreeNode.class, child);
		if (treeChild == null) {
			return -1;
		}
		final Object childObject = treeChild.lock();
		final Object parentObject = lock();
		if (childObject == null || parentObject == null) {
			return -1;
		}
		final AccessInterface access = tree.prepareAccess(param).select(parentObject);

		final int count = access.getCompositeParamCount();
		for (int i = 0; i < count; ++i) {
			Object c = access.get(access.getCompositeParam(i), Object.class);
			if (c == childObject) {
				return i;
			}
		}
		log.debug(child + " not found in " + this);
		return -1;
	}

	/**
	 * @return the param
	 */
	public CompositeParam getParam() {
		return param;
	}

	/**
	 * @return the tree
	 */
	public Tree getTree() {
		return tree;
	}


	@Override
	public String toString() {
		return param.toString();
	}

	public static Node tryGetNode(TreePath treePath) {
		return Casting.throwCast(TreeNode.class, treePath.getLastPathComponent()).tryCreateNode();
	}

	@Override
	public boolean isLeaf() {
		Object referent = lock();
		if (referent == null) {
			return true;
		}
		AccessInterface access = tree.prepareAccess(param).select(referent);
		return access.getCompositeParamCount() == 0;
	}


	@Override
	public void render(TreeCellRenderer renderer) {
		String name = param.getId();

		Object child = lock();
		if (child != null) {
			AccessInterface access = tree.prepareAccess(param).select(child);

			StringParam nameParam = Casting.tryCast(StringParam.class, access.getParam("name"));

			if (nameParam != null) {
				name = access.get(nameParam, String.class);
			}

			renderer.setToolTipText(new TooltipConstructor()
				.append("frams", access.getId())
				.append("java", child.getClass().getCanonicalName())
				.append("access", access.getClass().getSimpleName())
				.append("name", name)
				.append("id", param.getId())
				.append("object", Integer.toHexString(System.identityHashCode(child)))
				.append("number", number)
				.build());
		} else {
			renderer.setToolTipText("?");
		}
		renderer.setText(name);
		renderer.setIcon(ImageProvider.loadImage(TreeCellRenderer.findIconName(param)));

	}

}
